#include <string.h> // memset-funktiota varten

#include <msp430g2231.h>
// http://www.hutasu.net/Mikrokontrollerit/MCU_launchpad_osa25.htm
// Kalibrointiohjelman koko tavuina (byte) Flash/RAM = 436/2 + 32 bytea INFOC-muistia
// Ohjelmamuistin koko tavuina (byte) Flash/RAM = 574/24 + 32 bytea INFOC-muistia

// Launchpad mrityksi
#define LED_RED                      (BIT0)
#define LED_GREEN                   (BIT6)
#define INIT_LAUNCHPAD_LEDS  (P1DIR |= LED_RED+LED_GREEN)
#define LED_OFF                       (P1OUT &= ~(LED_RED + LED_GREEN))
#define LED_RED_ON					(P1OUT |= LED_RED)

// 7-segmentti mritykset
#define SEGMENTPORT			(P1OUT)
#define SEGMENTDIR			(P1DIR)
#define SEGMENTPINS			(BIT7 + BIT6 + BIT5 + BIT4)

// Matriisinppimistn mritykset

// MAX_KEYBOARD_SYMBOLS mrittelee montako nppint on matriisissa
#define MAX_KEYBOARD_SYMBOLS	16

// kun kalibroidaan, tytyy CALIBRATION olla pois kommentista. Eli kun ensimmist
// kertaa knnt ja ajat ohjelman, aseta alla oleva rivi pois kommentista ja aseta
// pohjelman breakpointti siell mainitulle paikalle.

#define CALIBRATION	// poista kommentista kun kalibroit!!!


// Kalibroidut arvot tallennetaan flash-muistiin,
// thn EI tarvitse koskea jos ei halua tallentaa muualle. 
#define CALIBRATION_ADDRESS			0x1040	// Segmentti C alkuarvo


// Tt voi muuttaa ja kokeilla eri arvoja,
// mutta tmn arvon ei kannata olla kovin suuri.
#define DEFAULT_NOISE_THRESHOLD		2		// ns. kohina-arvo AD-muuntimelle


// #ifdef <avainSana> komennolla voidaan ohjata kntjn tyt. Kun avainsana CALIBRATION on kommentoitu, niin
// #ifdef ja #endif avainsanojen vlist lohkoa ei knnet ollenkaan konekielelle.
#ifdef CALIBRATION
	volatile unsigned int keyValues[MAX_KEYBOARD_SYMBOLS];
#endif
#ifndef CALIBRATION
	// Jos kytss on esim. 3x4 nppimist, tai jos kytt kustomoitua nppimist
	// niin muuta alla olevaa "nppinkarttaa" vastaamaan sit.
	unsigned char symbols[MAX_KEYBOARD_SYMBOLS] = {
									1,  2,  3,   'F',
									4,  5,  6,   'E',
									7,  8,  9,   'D',
									'A',0, 'B',  'C'};
#endif

// Launchpad funktiot
inline void InitLaunchpad();
// Watchdog pysytysfunktio
inline void StopWatchdog();

// 7-Segmenttifunktiot
void InitSegmentDriver();
void SetSegmentNumber(unsigned char n);

// Erillisi matriisinppinfunktioita ei ole
// vaan kytetn nppimen lukemiseen sek ADC:n
// ett Flashin funktioita.

// ADC:n kyttfunktiot
void InitAdcDriver();
void StartAdc();
void WaitAdc();
unsigned int ReadAdcValue();

// Flash tallennus- ja lukufunktiot
#ifdef CALIBRATION
	void WriteCalibrationDataToFlash();
#endif
#ifndef CALIBRATION
	unsigned char ReadKeyDataFromFlash();
#endif

void main(void)
{
	#ifdef CALIBRATION
		unsigned char index = 0;
		unsigned int adcValue = 0;
		// C:n standardifunktio memset alustaa muistialueen haluttuun arvoon. parametrit: muistialue, arvo, pituus.
		memset((void *)keyValues,0,sizeof(keyValues)); // alustetaan taulukko nollaan
	#endif
	#ifndef CALIBRATION
		unsigned char key = 0;
	#endif

     StopWatchdog();
     InitLaunchpad();
     InitSegmentDriver();
     InitAdcDriver();
     while(1)
     {
		#ifdef CALIBRATION
    	 	 for(index = 0; index < MAX_KEYBOARD_SYMBOLS; index++)
    	 	 {
    	 		 // painetaan nappeja jrjestyksess kunnes
    	 		 // kaikki napit on kalibroitu
    	 		 // aseta breakpoint alle ja paina nappeja
    	 		 // jrjestyksess 1,2,3,F, sitten seuraava rivi jne...
    	 		 __no_operation();  // <-- breakpoint thn
    	 		 StartAdc();
				 WaitAdc();
				 adcValue = ReadAdcValue();
				 // estetn painamattomien nappien tallennus
				 if(adcValue > 5)
				 {
					 keyValues[index] = adcValue;
				 }
				 else
				 {
					 index--;
				 }
    	 	 }
    	 	 // aseta toinen breakpoint alle jolloin voit
    	 	 // viel keskeytt flash tallennuksen keskeyttmll debug-session
    	 	__no_operation();

    	 	// Tallennetaan tiedot flashiin, steppaa tm F6-napilla yli tai
    	 	// aja uudestaan ensimmiseen breakpointtiin. Kun tiedot on tallennettu, voit lopettaa session.
    	 	// Tietojen tallennuksen net Memory Browser ikkunasta kirjoittamalla "Enter Location Here" -laatikkoon 0x1040,
    	 	// mikli kyseisess sijainnissa on muuta kuin 0xFF tavuja, niin tiedot on tallennettu onnistuneesti.
    	 	WriteCalibrationDataToFlash();
    	 	while(1); // pysytetn CPU
    	 	// tutki memory-ikkunassa tallentuiko tiedot oikein
		#endif
		#ifndef CALIBRATION
    	 	// Tm osa suoritetaan ns. normaalissa tilassa. Muista clean-rebuild!
    	 	StartAdc();
    	 	WaitAdc();
    	 	// Funktiokutsussa parametrina voi kytt mys toista funktiota, mutta tllin
    	 	// kannattaa tarkastaa ett parametrifunktio palauttaa saman tyyppisen arvon
    	 	// kuin mik parametrin pit olla (unsigned int <-> unsigned int tss tapauksessa)
			key = ReadKeyDataFromFlash( ReadAdcValue() );
			if(key != 0xFF)
			{
				// asetetaan numero nytlle
				SetSegmentNumber(key);
				// Jtetn numero sekunniksi plle
				__delay_cycles(1000000);
			}
			else
			{
			 SetSegmentNumber(0xFF);
			}
    	#endif
     }
}

// Launchpad alustusfunktio
inline void InitLaunchpad()
{
     INIT_LAUNCHPAD_LEDS; // Ledien alustus
     LED_OFF;  // LED1 ja LED2 pois plt
     P1DIR = 0xFF;
     P1IE = 0;
     P1IES = 0;
     P1REN = 0;
     P1OUT = 0;
     // alustetaan CPU:n kello 1 Mhz
     BCSCTL1 = CALBC1_1MHZ;
     DCOCTL = CALDCO_1MHZ;
     // BCSCTL3 = XCAP_3;
     // kiteen alustus on kommentoitu koska sit ei nyt tarvita
}

// Watchdog pysytysfunktio
inline void StopWatchdog()
{
	WDTCTL = WDTPW + WDTHOLD;    // watchdog pysytys
}
// Segmentin alustusfunktio
void InitSegmentDriver()
{
	// asetetaan pinnit lhdksi ja nollataan kaikki
	SEGMENTDIR |= SEGMENTPINS;
	// asetetaan segmenttipinnit "1" tilaan, jolloin nytss
	// ei ny mitn. Jos porttipinnit on "0" niin 0 nkyy nytss.
	SEGMENTPORT |= SEGMENTPINS;
}
// Segmentin numeron asetusfunktio. Parametrina haluttu numero (0...9 mahdolliset nytt).
void SetSegmentNumber(unsigned char n)
{
	n <<=4;
	// Koska segmentti on kytketty P1.4-P1.7 nastoihin, niin
	// tss tarvitaan vain portin P1 maskaus
	SEGMENTPORT = 0xF0 & n;
}

void InitAdcDriver()
{
	// Enabloidaan input pinni
	ADC10AE0 = BIT3;
	// Asetetaan input A3 + MCLK
	ADC10CTL1 = INCH_3 + ADC10SSEL_2;
	// Vcc, GND referenssin, 4 x ADC10CLKS, ADC10 plle
	ADC10CTL0 = SREF_0 + ADC10SHT_0 + ADC10ON;
	// MCLK on 1 Mhz ja sample timerin pienin arvo on 4x, niin saadaan 4 us:n nytteistysaika
}
void StartAdc()
{
	// kynnistetn muunnos
	ADC10CTL0 |= ENC + ADC10SC;
}
void WaitAdc()
{
	// odotetaan muunnoksen valmistumista
	while (ADC10CTL1 & ADC10BUSY);
	// mahdollinen paikka jumittumiselle ikuiseen silmukkaan...
}
unsigned int ReadAdcValue()
{
	return ADC10MEM;
}

#ifdef CALIBRATION
// Flash tallennusfunktio
void WriteCalibrationDataToFlash()
{
	unsigned int *Flash_ptr;
	unsigned char i = 0;

	// Ladataan muistipaikka
	Flash_ptr = (unsigned int *)CALIBRATION_ADDRESS;

	// Pyyhitn lohko ja poistetaan lukko
	FCTL1 = FWKEY + ERASE;
	FCTL3 = FWKEY;
	*Flash_ptr = 0;
	while ( ( FCTL3 & BUSY )  !=  0 );
	FCTL1 = FWKEY + WRT;

	// kopioidaan kalibrointitiedot
	for(i = 0; i < MAX_KEYBOARD_SYMBOLS; i++)
	{
		// tallennetaan kaikki MAX_KEYBOARD_SYMBOLS (16) nppint
		*Flash_ptr++ = (unsigned int) keyValues[i];
		while ( ( FCTL3 & BUSY )  !=  0 );
	}

	// Lukitaan flash-muisti
	FCTL1 = FWKEY;
	FCTL3 = FWKEY + LOCK;
}
#endif
#ifndef CALIBRATION
// Flash lukufunktio
unsigned char ReadKeyDataFromFlash(unsigned int adcValue)
{
	unsigned int *Flash_ptr;
	int tempValue=0;
	unsigned char valueFound = 0;
	unsigned char i = 0;

	// ladataan muistipaikka
	Flash_ptr = (unsigned int *)CALIBRATION_ADDRESS;

	// adcValue on kalibroiduista tiedoista etsittv arvo
	while(i++ < MAX_KEYBOARD_SYMBOLS)
	{
		// tutkitaan lytyyk toimitettua ADC-arvoa
		// kalibrointitiedoista tietyn kohinarajan sisll

		// ladataan ensimminen arvo
		tempValue = *Flash_ptr++;

		// tutkitaan onko rajoissa. tehdn ensin tutkinta kumpi on suurempi
		if(tempValue > adcValue)
		{
			// ladattu kalib. arvo on suurempi, esimerkiksi adcValue = 531 ja tempValue = 64
			tempValue = adcValue - tempValue;
			if((tempValue > 0) && (tempValue < DEFAULT_NOISE_THRESHOLD))
			{
				// rajoissa ja lytyi
				valueFound = 1;
				break;
			}
			// muutoin ei lytynyt
		}
		else if(tempValue < adcValue)
		{
			// ladattu kalib. arvo on pienempi
			tempValue = tempValue - adcValue;
			if((tempValue > 0) && (tempValue < DEFAULT_NOISE_THRESHOLD))
			{
				// rajoissa ja lytyi
				valueFound = 1;
				break;
			}
			// muutoin ei lytynyt
		}
		else
		{
			// tasan yhtsuuri koska ei voi olla muuta joten arvo lytyi suoraan
			valueFound = 1;
			break;
		}
	}

	// Lopuksi tutkitaan lytyik lukua taulukosta ollenkaan
	if(valueFound)
	{
		// luku oli lytynyt, palautetaan indeksipaikan mukainen symboli
		return symbols[i-1];
	}
	else
	{
		return 0xFF;
	}
}
#endif
